#include "m_pd.h"
#include <math.h>
#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif

/* ------------------------ sof~ ----------------------------- */

/* sof~ - first order filters, lowpass and highpass. */

static t_class *sof_class;

typedef struct _sof
{
    t_object x_obj; 	/* obligatory header */
    t_float frequency;  /* place to hold inlet's value if it's set by message */
    t_float frequencyold;
    t_float bandwidth;
    t_float samplerate;
    t_float PI;
    t_float c;
    t_float d;
    t_int type;  // low or highpass
    t_float in1;
    t_float out;
    t_float in2;
    t_float out1;
    t_float out2;
} t_sof;


void sof_bandwidth(t_sof *x, t_floatarg g)
{
    t_float tf, t2f;
    x->bandwidth = g;
    tf = tanf(x->PI * x->bandwidth/x->samplerate);
    t2f = tanf(x->PI * 2.0f * x->bandwidth/x->samplerate);
    x->c = (tf - 1.0f)/(tf + 1.0f);
    
}
void sof_notch(t_sof *x)
{
    x->type = 0;
}

void sof_bandpass(t_sof *x)
{
    x->type = 1;
}

void sof_allpass(t_sof *x)
{
    x->type = 2;
}

void sof_bypass(t_sof *x)
{
    x->type = 3;
}

/* this is the actual performance routine which acts on the samples.
    It's called with a single pointer "w" which is our location in the
    DSP call list.  We return a new "w" which will point to the next item
    after us.  Meanwhile, w[0] is just a pointer to dsp-perform itself
    (no use to us), w[1] and w[2] are the input and output vector locations,
    and w[3] is the number of points to calculate. */
static t_int *sof_perform(t_int *w)
{
    t_float tf;
    
    t_sof *x = (t_sof *)(w[1]);
    t_float *in = (t_float *)(w[2]);
    t_float *out = (t_float *)(w[3]);
    int n = (int)(w[4]);
    
    // calculate coefficient only if frequency changes
    if(x->frequency != x->frequencyold)
    {
        if(x->frequency > 20000.0f)
            x->frequencyold = x->frequency = 20000.0f;
        else if(x->frequency < 0.0001f)
            x->frequencyold = x->frequency = 0.0001f;
        else
            x->frequencyold = x->frequency;

        x->d = -1.0f * cosf(2.0f * x->PI * x->frequency/x->samplerate);
    }
    
    while (n--)
    {
    	float input = *(in++);
        x->out = x->c * (x->out2 - input)
            + x->d * (1.0f - x->c) * (x->in1 - x->out1)
            + x->in2;
        x->out2 = x->out1;
        x->out1 = x->out;
        x->in2 = x->in1;
        x->in1 = input;
        switch(x->type)
        {
            case 0:
                *out++ = (input + x->out) * 0.5f;
                break;
            case 1:
                *out++ = (input - x->out) * 0.5f;
                break;
            case 2:
                *out++ = x->out;
                break;
            case 3:
                *out++ = input;
            default:
                break;
        }
    }
    return (w+5);
}

    /* called to start DSP.  Here we call Pd back to add our perform
    routine to a linear callback list which Pd in turn calls to grind
    out the samples. */
static void sof_dsp(t_sof *x, t_signal **sp)
{
    x->samplerate = sp[0]->s_sr;
    dsp_add(sof_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}

static void *sof_new(void)
{
    t_sof *x = (t_sof *)pd_new(sof_class);
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("float"), gensym("bandwidth"));
    outlet_new(&x->x_obj, gensym("signal"));
    // initialize state variables
    x->frequency = x->frequencyold = x->c = x->d = 0.0f;
    x->PI = 4.0f * atanf(1.0f);
    x->samplerate = 44100.0f;
    x->type = 1;
    x->in1 = x->out1 = 0.0f;
    x->in2 = x->out2 = 0.0f;
    return (x);
}

    /* this routine, which must have exactly this name (with the "~" replaced
    by "_tilde) is called when the code is first loaded, and tells Pd how
    to build the "class". */
void sof_tilde_setup(void)
{
    sof_class = class_new(gensym("sof~"), (t_newmethod)sof_new, 0,
    	sizeof(t_sof), 0, A_DEFFLOAT, 0);
	    /* this is magic to declare that the leftmost, "main" inlet
	    takes signals; other signal inlets are done differently... */
    CLASS_MAINSIGNALIN(sof_class, t_sof, frequency);
    class_addmethod(sof_class, (t_method)sof_bandwidth, gensym("bandwidth"), A_FLOAT, 0);
    class_addmethod(sof_class, (t_method)sof_notch, gensym("notch"), 0);
    class_addmethod(sof_class, (t_method)sof_bandpass, gensym("bandpass"), 0);
    class_addmethod(sof_class, (t_method)sof_allpass, gensym("allpass"), 0);
    class_addmethod(sof_class, (t_method)sof_bypass, gensym("bypass"), 0);
    class_addmethod(sof_class, (t_method)sof_dsp, gensym("dsp"), 0);
}
